----------------------------------------------------------------------------------
--                                 _             _
--                                | |_  ___ _ __(_)__ _
--                                | ' \/ -_) '_ \ / _` |
--                                |_||_\___| .__/_\__,_|
--                                         |_|
--
----------------------------------------------------------------------------------
--
-- Company: hepia
-- Author: Joachim Schmidt <joachim.schmidt@hesge.ch
--
-- Module Name: reset_logic - rtl
-- Target Device: SCALP xc7z015clg485-2
-- Tool version: 2020.2
-- Description: Aurora reset logic module
--
-- Last update: 2020-09-28
--
---------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library UNISIM;
use UNISIM.VCOMPONENTS.all;

entity reset_logic is

    port (
        UserClkxCI     : in  std_logic;
        InitClkxCI     : in  std_logic;
        ResetxRI       : in  std_logic;
        GTResetxRI     : in  std_logic;
        SystemResetxRO : out std_logic;
        GTResetxRO     : out std_logic);

end entity reset_logic;

architecture behavioral of reset_logic is

    -- Constants
    ---------------------------------------------------------------------------
    constant C_CDC_TYPE     : integer range 0 to 2  := 1;
    constant C_RESET_STATE  : integer range 0 to 1  := 0;
    constant C_SINGLE_BIT   : integer range 0 to 1  := 1;
    constant C_FLOP_INPUT   : integer range 0 to 1  := 1;
    constant C_VECTOR_WIDTH : integer range 0 to 32 := 2;
    constant C_MTBF_STAGES  : integer range 0 to 6  := 5;

    -- Components
    ---------------------------------------------------------------------------
    component cdc_sync is
        generic (
            C_CDC_TYPE     : integer range 0 to 2;
            C_RESET_STATE  : integer range 0 to 1;
            C_SINGLE_BIT   : integer range 0 to 1;
            C_FLOP_INPUT   : integer range 0 to 1;
            C_VECTOR_WIDTH : integer range 0 to 32;
            C_MTBF_STAGES  : integer range 0 to 6);
        port (
            PrimaryClkxCAI     : in  std_logic;
            PrimaryResetxRNI   : in  std_logic;
            PrimaryxSI         : in  std_logic;
            PrimaryxDI         : in  std_logic_vector((C_VECTOR_WIDTH - 1) downto 0);
            PrimaryAckxSO      : out std_logic;
            SecondaryClkxCAI   : in  std_logic;
            SecondaryResetxRNI : in  std_logic;
            SecondaryxSO       : out std_logic;
            SecondaryxDO       : out std_logic_vector((C_VECTOR_WIDTH - 1) downto 0));
    end component cdc_sync;

    -- Signals
    ---------------------------------------------------------------------------    
    -- Internal reset signals
    signal GTPResetxR                             : std_logic                    := '0';
    signal GTPResetSyncxR                         : std_logic                    := '0';
    signal DebounceGTPResetxD                     : std_logic_vector(3 downto 0) := (others => '0');
    signal DebounceResetxD                        : std_logic_vector(3 downto 0) := (others => '0');
    signal DebounceResetxS                        : std_logic                    := '1';
    -- Attributes    
    attribute async_reg                           : string;
    attribute shift_extract                       : string;
    attribute async_reg of DebounceGTPResetxD     : signal is "true";
    attribute shift_extract of DebounceGTPResetxD : signal is "no";
    attribute async_reg of DebounceResetxD        : signal is "true";
    attribute shift_extract of DebounceResetxD    : signal is "no";

begin  -- architecture behavioral

    -- Asynchronous statements

    GTResetOutxAS  : GTResetxRO     <= GTPResetxR;
    SystemResetxAS : SystemResetxRO <= DebounceResetxS;

    -- GTP Reset sync.
    ---------------------------------------------------------------------------
    GTPResetSyncxI : entity work.cdc_sync
        generic map (
            C_CDC_TYPE     => C_CDC_TYPE,
            C_RESET_STATE  => C_RESET_STATE,
            C_SINGLE_BIT   => C_SINGLE_BIT,
            C_FLOP_INPUT   => C_FLOP_INPUT,
            C_VECTOR_WIDTH => C_VECTOR_WIDTH,
            C_MTBF_STAGES  => C_MTBF_STAGES)
        port map (
            PrimaryClkxCAI     => InitClkxCI,
            PrimaryResetxRNI   => '1',
            PrimaryxSI         => GTPResetxR,
            PrimaryxDI         => "00",
            PrimaryAckxSO      => open,
            SecondaryClkxCAI   => UserClkxCI,
            SecondaryResetxRNI => '1',
            SecondaryxSO       => GTPResetSyncxR,
            SecondaryxDO       => open);

    -- Synchronous statements

    -- Debounce the GTResetInxR using the InitClkxCI
    ---------------------------------------------------------------------------
    DebounceGTPResetxP : process (InitClkxCI) is
    begin  -- process DebounceGTPResetxP
        if rising_edge(InitClkxCI) then
            DebounceGTPResetxD <= DebounceGTPResetxD(2 downto 0) & GTResetxRI;
            GTPResetxR         <= DebounceGTPResetxD(3) and
                          DebounceGTPResetxD(2) and
                          DebounceGTPResetxD(1) and
                          DebounceGTPResetxD(0);
        end if;
    end process DebounceGTPResetxP;

    -- Debounce the ResetxR signal
    ---------------------------------------------------------------------------
    DebounceResetStp1xP : process (GTPResetSyncxR, UserClkxCI) is
    begin  -- process DebounceResetStp1xP
        if GTPResetSyncxR = '1' then
            DebounceResetxD <= (others => '1');
        elsif rising_edge(UserClkxCI) then
            DebounceResetxD <= DebounceResetxD(2 downto 0) & ResetxRI;
        end if;
    end process DebounceResetStp1xP;

    DebounceResetStp2xP : process (UserClkxCI) is
    begin  -- process DebounceResetStp2xP
        if rising_edge(UserClkxCI) then
            DebounceResetxS <= DebounceResetxD(3) and
                               DebounceResetxD(2) and
                               DebounceResetxD(1) and
                               DebounceResetxD(0);
        end if;
    end process DebounceResetStp2xP;

end architecture behavioral;
