library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library unisim;
use unisim.vcomponents.all;

entity scalp_safe_firmware_reg_bank is

    generic (
        C_AXI4_ARADDR_SIZE : integer range 0 to 32 := 32;
        C_AXI4_RDATA_SIZE  : integer range 0 to 32 := 32;
        C_AXI4_RRESP_SIZE  : integer range 0 to 2  := 2;
        C_AXI4_AWADDR_SIZE : integer range 0 to 32 := 32;
        C_AXI4_WDATA_SIZE  : integer range 0 to 32 := 32;
        C_AXI4_WSTRB_SIZE  : integer range 0 to 4  := 4;
        C_AXI4_BRESP_SIZE  : integer range 0 to 2  := 2;
        C_AXI4_ADDR_SIZE   : integer range 0 to 32 := 12;
        C_AXI4_DATA_SIZE   : integer range 0 to 32 := 32);

    port (
        -- Clock and reset
        SAxiClkxCI         : in  std_ulogic;
        SAxiRstxRANI       : in  std_ulogic;
        -- AXI4 Lite
        -- Read Channel
        -- Read Address Channel
        SAxiARAddrxDI      : in  std_ulogic_vector((C_AXI4_ARADDR_SIZE - 1) downto 0);
        SAxiARValidxSI     : in  std_ulogic;
        SAxiARReadyxSO     : out std_ulogic;
        -- Read Data Channel
        SAxiRDataxDO       : out std_ulogic_vector((C_AXI4_RDATA_SIZE - 1) downto 0);
        SAxiRRespxDO       : out std_ulogic_vector((C_AXI4_RRESP_SIZE - 1) downto 0);
        SAxiRValidxSO      : out std_ulogic;
        SAxiRReadyxSI      : in  std_ulogic;
        -- Write Channel
        -- Write Address Channel
        SAxiAWAddrxDI      : in  std_ulogic_vector((C_AXI4_AWADDR_SIZE - 1) downto 0);
        SAxiAWValidxSI     : in  std_ulogic;
        SAxiAWReadyxSO     : out std_ulogic;
        -- Write Data Channel
        SAxiWDataxDI       : in  std_ulogic_vector((C_AXI4_WDATA_SIZE - 1) downto 0);
        SAxiWStrbxDI       : in  std_ulogic_vector((C_AXI4_WSTRB_SIZE - 1) downto 0);
        SAxiWValidxSI      : in  std_ulogic;
        SAxiWReadyxSO      : out std_ulogic;
        -- Write Response Channel
        SAxiBRespxDO       : out std_ulogic_vector((C_AXI4_BRESP_SIZE - 1) downto 0);
        SAxiBValidxSO      : out std_ulogic;
        SAxiBReadyxSI      : in  std_ulogic;
        -- Register Access IO
        RgbLedsCtrlPortxDO : out std_ulogic_vector((C_AXI4_DATA_SIZE - 1) downto 0));

end scalp_safe_firmware_reg_bank;

architecture behavioral of scalp_safe_firmware_reg_bank is

    -- Constants
    constant C_AXI4_RRESP_OKAY   : std_ulogic_vector((C_AXI4_RRESP_SIZE - 1) downto 0) := "00";
    constant C_AXI4_RRESP_EXOKAY : std_ulogic_vector((C_AXI4_RRESP_SIZE - 1) downto 0) := "01";
    constant C_AXI4_RRESP_SLVERR : std_ulogic_vector((C_AXI4_RRESP_SIZE - 1) downto 0) := "10";
    constant C_AXI4_RRESP_DECERR : std_ulogic_vector((C_AXI4_RRESP_SIZE - 1) downto 0) := "11";
    constant C_AXI4_BRESP_OKAY   : std_logic_vector((C_AXI4_BRESP_SIZE - 1) downto 0)  := "00";
    constant C_AXI4_BRESP_EXOKAY : std_logic_vector((C_AXI4_BRESP_SIZE - 1) downto 0)  := "01";
    constant C_AXI4_BRESP_SLVERR : std_logic_vector((C_AXI4_BRESP_SIZE - 1) downto 0)  := "10";
    constant C_AXI4_BRESP_DECERR : std_logic_vector((C_AXI4_BRESP_SIZE - 1) downto 0)  := "11";

    -- Signals
    -- Clock and reset
    signal SAxiClkxC          : std_ulogic                                         := '0';
    signal SAxiRstxRAN        : std_ulogic                                         := '0';
    -- AXI4 Lite
    signal SAxiARReadyxS      : std_ulogic                                         := '0';
    signal SAxiRValidxS       : std_ulogic                                         := '0';
    signal SAxiBValidxS       : std_ulogic                                         := '0';
    signal SAxiWReadyxS       : std_ulogic                                         := '0';
    signal SAxiAWReadyxS      : std_ulogic                                         := '0';
    signal WrAddrxDN          : std_ulogic_vector((C_AXI4_ADDR_SIZE - 1) downto 0) := (others => '0');
    signal WrAddrxDP          : std_ulogic_vector((C_AXI4_ADDR_SIZE - 1) downto 0) := (others => '0');
    -- Signals of access to the register bank
    signal RdAddrxD           : std_ulogic_vector((C_AXI4_ADDR_SIZE - 1) downto 0) := (others => '0');
    signal RdDataxD           : std_ulogic_vector((C_AXI4_DATA_SIZE - 1) downto 0) := (others => '0');
    signal RdValidxS          : std_ulogic                                         := '0';
    signal WrAddrxD           : std_ulogic_vector((C_AXI4_ADDR_SIZE - 1) downto 0) := (others => '0');
    signal WrDataxD           : std_ulogic_vector((C_AXI4_DATA_SIZE - 1) downto 0) := (others => '0');
    signal WrValidxS          : std_ulogic                                         := '0';
    -- Registers list
    signal RgbLedsCtrlPortxDN : std_ulogic_vector((C_AXI4_DATA_SIZE - 1) downto 0) := x"00000000";
    signal RgbLedsCtrlPortxDP : std_ulogic_vector((C_AXI4_DATA_SIZE - 1) downto 0) := x"00000000";

    -- Attributes
    attribute mark_debug                       : string;
    attribute keep                             : string;
    --
    attribute mark_debug of RdAddrxD           : signal is "true";
    attribute keep of RdAddrxD                 : signal is "true";
    attribute mark_debug of RdDataxD           : signal is "true";
    attribute keep of RdDataxD                 : signal is "true";
    attribute mark_debug of RdValidxS          : signal is "true";
    attribute keep of RdValidxS                : signal is "true";
    attribute mark_debug of WrAddrxD           : signal is "true";
    attribute keep of WrAddrxD                 : signal is "true";
    attribute mark_debug of WrDataxD           : signal is "true";
    attribute keep of WrDataxD                 : signal is "true";
    attribute mark_debug of WrValidxS          : signal is "true";
    attribute keep of WrValidxS                : signal is "true";
    attribute mark_debug of RgbLedsCtrlPortxDP : signal is "true";
    attribute keep of RgbLedsCtrlPortxDP       : signal is "true";

begin

    assert C_AXI4_RDATA_SIZE = C_AXI4_DATA_SIZE
        report "RDATA and DATA vectors must be the same" severity failure;

    assert C_AXI4_ARADDR_SIZE >= C_AXI4_ADDR_SIZE
        report "ARADDR and ADDR vectors must be the same" severity failure;

    assert C_AXI4_WDATA_SIZE = C_AXI4_DATA_SIZE
        report "WDATA and DATA vectors must be the same" severity failure;

    assert C_AXI4_AWADDR_SIZE >= C_AXI4_ADDR_SIZE
        report "AWADDR and ADDR vectors must be the same" severity failure;

    EntityIOxB : block is
    begin  -- block EntityIOxB

        -- Clock and reset
        SAxiClkxAS     : SAxiClkxC                                  <= SAxiClkxCI;
        SAxiRstxAS     : SAxiRstxRAN                                <= SAxiRstxRANI;
        -- Read Channel
        SAxiARReadyxAS : SAxiARReadyxSO                             <= SAxiARReadyxS;
        SAxiRValidxAS  : SAxiRValidxSO                              <= SAxiRValidxS;
        SAxiRDataxAS   : SAxiRDataxDO                               <= RdDataxD;
        RdValidxAS     : RdValidxS                                  <= SAxiARValidxSI;
        RdAddrxAS      : RdAddrxD((C_AXI4_ADDR_SIZE - 1) downto 0)  <= SAxiARAddrxDI((C_AXI4_ADDR_SIZE - 1) downto 0);
        SAxiRRespxAS   : SAxiRRespxDO                               <= C_AXI4_RRESP_OKAY;
        -- Write Channel
        SAxiBRespxAS   : SAxiBRespxDO                               <= C_AXI4_BRESP_OKAY;
        SAxiBValidxAS  : SAxiBValidxSO                              <= SAxiBValidxS;
        SAxiWReadyxAS  : SAxiWReadyxSO                              <= SAxiWReadyxS;
        SAxiAWReadyxAS : SAxiAWReadyxSO                             <= SAxiAWReadyxS;
        WrValidxAS     : WrValidxS                                  <= SAxiWValidxSI;
        WrDataxAS      : WrDataxD                                   <= SAxiWDataxDI;
        WrAddrOutxAS   : WrAddrxD                                   <= WrAddrxDP;
        WrAddrxAS      : WrAddrxDN((C_AXI4_ADDR_SIZE - 1) downto 0) <= SAxiAWAddrxDI((C_AXI4_ADDR_SIZE - 1) downto 0) when
                                                                  SAxiAWValidxSI = '1' else
                                                                  WrAddrxDP((C_AXI4_ADDR_SIZE - 1) downto 0);
        RgbLedsCtrlPortxAS : RgbLedsCtrlPortxDO <= RgbLedsCtrlPortxDP;

    end block EntityIOxB;

    AXI4LitexB : block is
    begin  -- block AXI4LitexB

        ReadChannelxB : block is
        begin  -- block ReadChannelxB

            ReadAddrChanxP : process (SAxiClkxC, SAxiRstxRAN) is

                variable StateAfterResetxS : boolean := true;

            begin  -- process ReadAddrChanxP
                if SAxiRstxRAN = '0' then
                    SAxiARReadyxS     <= '0';
                    StateAfterResetxS := true;
                elsif rising_edge(SAxiClkxC) then
                    if StateAfterResetxS = true then
                        SAxiARReadyxS     <= '1';
                        StateAfterResetxS := false;
                    else
                        SAxiARReadyxS <= SAxiARReadyxS;
                    end if;

                    if SAxiARValidxSI = '1' then
                        SAxiARReadyxS <= '0';
                    end if;

                    if SAxiARReadyxS <= '0' and SAxiRReadyxSI = '1' then
                        SAxiARReadyxS <= '1';
                    end if;
                end if;
            end process ReadAddrChanxP;

            ReadDataChanxP : process (SAxiClkxC, SAxiRstxRAN) is
            begin  -- process ReadDataChanxP
                if SAxiRstxRAN = '0' then
                    SAxiRValidxS <= '0';
                elsif rising_edge(SAxiClkxC) then
                    SAxiRValidxS <= SAxiRValidxS;

                    if SAxiARValidxSI = '1' and SAxiARReadyxS = '1' then
                        SAxiRValidxS <= '1';
                    end if;

                    if SAxiRValidxS = '1' and SAxiRReadyxSI = '1' then
                        SAxiRValidxS <= '0';
                    end if;
                end if;
            end process ReadDataChanxP;

        end block ReadChannelxB;

        WriteChannelxB : block is
        begin  --block WriteChannelxB        

            WrAddrRegxP : process (SAxiClkxC, SAxiRstxRAN) is
            begin  -- process WrAddrRegxP
                if SAxiRstxRAN = '0' then
                    WrAddrxDP <= (others => '0');
                elsif rising_edge(SAxiClkxC) then
                    WrAddrxDP <= WrAddrxDN;
                end if;
            end process WrAddrRegxP;

            WriteAddrChanxP : process (SAxiClkxC, SAxiRstxRAN) is

                variable StateAfterResetxS : boolean := true;

            begin  -- process WriteAddrChanxP
                if SAxiRstxRAN = '0' then
                    SAxiAWReadyxS     <= '0';
                    StateAfterResetxS := true;
                elsif rising_edge(SAxiClkxC) then
                    if StateAfterResetxS = true then
                        SAxiAWReadyxS     <= '1';
                        StateAfterResetxS := false;
                    else
                        SAxiAWReadyxS <= SAxiAWReadyxS;
                    end if;

                    if SAxiAWValidxSI = '1' then
                        SAxiAWReadyxS <= '0';
                    end if;

                    if SAxiWValidxSI = '1' then
                        SAxiAWReadyxS <= '1';
                    end if;
                end if;
            end process WriteAddrChanxP;

            WriteDataChanxP : process (SAxiClkxC, SAxiRstxRAN) is
            begin  -- process WriteDataChanxP
                if SAxiRstxRAN = '0' then
                    SAxiWReadyxS <= '0';
                elsif rising_edge(SAxiClkxCI) then
                    SAxiWReadyxS <= SAxiWReadyxS;

                    if SAxiAWValidxSI = '1' and SAxiAWReadyxS = '1' then
                        SAxiWReadyxS <= '1';
                    end if;

                    if SAxiWValidxSI = '1' and SAxiWReadyxS = '1' then
                        SAxiWReadyxS <= '0';
                    end if;
                end if;
            end process WriteDataChanxP;

            WriteRespChanxP : process (SAxiClkxC, SAxiRstxRAN) is
            begin  -- process WriteRespChanxP
                if SAxiRstxRAN = '0' then
                    SAxiBValidxS <= '0';
                elsif rising_edge(SAxiClkxC) then
                    SAxiBValidxS <= SAxiBValidxS;

                    if SAxiWValidxSI = '1' and SAxiWReadyxS = '1' then
                        SAxiBValidxS <= '1';
                    end if;

                    if SAxiBValidxS = '1' and SAxiBReadyxSI = '1' then
                        SAxiBValidxS <= '0';
                    end if;
                end if;
            end process WriteRespChanxP;

        end block WriteChannelxB;

    end block AXI4LitexB;

    ScalpSafeFirmwareRegBankxB : block is
    begin  -- block ScalpSafeFirmwareRegBankxB

        WriteRegPortxP : process (WrAddrxD, WrDataxD, WrValidxS,
                                  RgbLedsCtrlPortxDP) is
        begin  -- process WriteRegPortxP
            RgbLedsCtrlPortxDN <= RgbLedsCtrlPortxDP;

            if WrValidxS = '1' then
                case WrAddrxD is
                    when x"000" => RgbLedsCtrlPortxDN <= WrDataxD;
                    when x"004" => RgbLedsCtrlPortxDN <= RgbLedsCtrlPortxDP or WrDataxD;
                    when x"008" => RgbLedsCtrlPortxDN <= RgbLedsCtrlPortxDP and not WrDataxD;

                    when others => null;
                end case;
            end if;
        end process WriteRegPortxP;

        ReadRegPortxP : process (SAxiClkxC, SAxiRstxRAN) is
        begin  -- process ReadRegPortxP
            if SAxiRstxRAN = '0' then
                RdDataxD <= (others => '0');
            elsif rising_edge(SAxiClkxC) then
                RdDataxD <= (others => '0');

                if RdValidxS = '1' then
                    case RdAddrxD is
                        when x"000" => RdDataxD <= RgbLedsCtrlPortxDP;

                        when others => RdDataxD <= (others => '0');
                    end case;
                end if;
            end if;
        end process ReadRegPortxP;

        UpdateRegBankxP : process (SAxiClkxC, SAxiRstxRAN) is
        begin  -- process UpdateRegBankxP
            if SAxiRstxRAN = '0' then
                RgbLedsCtrlPortxDP <= x"00000000";

            elsif rising_edge(SAxiClkxC) then
                RgbLedsCtrlPortxDP <= RgbLedsCtrlPortxDN;

            end if;
        end process UpdateRegBankxP;

    end block ScalpSafeFirmwareRegBankxB;

end behavioral;
